# SPDX-License-Identifier: Apache-2.0

# To register shape calculators for Core ML neural
# network operators, import associated modules here.
from . import BatchNorm
from . import BidirectionalLSTM
from . import Concat
from . import Convolution
from . import Crop
from . import Dot
from . import Embed
from . import Flatten
from . import GRU
from . import IdentityFloat
from . import InnerProduct
from . import LoadConstant
from . import LoadConstantND
from . import LSTM
from . import Merge
from . import Pad
from . import Permute
from . import Pool
from . import Reduce
from . import ReorganizeData
from . import Reshape
from . import ReshapeStatic
from . import SequenceRepeat
from . import Slice
from . import Split
from . import Upsample
